/** @file leafnode.h */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#ifndef _LEAF_NODE_H_
#define _LEAF_NODE_H_
#include "mempool.h"
#include "simplehash.h"
#include <slist>

/// Encapsulates a WebNode which exists on another machine.
class LeafNode; // defined below
/// All LeafNodes are allocated on the heap as LeafNodePtrs.
typedef LeafNode* LeafNodePtr;

/// Allows a leaf node to be found through an url
typedef SimpleHashTable<LeafNodePtr> SimpleLeafNodePtrHashTable;

/// A linked list of all LeafNode objects
typedef slist<LeafNodePtr> LeafNodePtrList;

#define Mb 1048576L
#define LEAFNODE_MEMPOOL_DELTA ((1 * Mb)/sizeof(LeafNode))

/// Contains all the information about a web document stored somewhere else.
struct LeafNodeStruct {
  // the ordering here is important due to alignment issues
  uint32 id;
  uint16 date;
  uint32 occupation_count;
};

/// Encapsulates a web document on another machine.
/**
 * When several machines process (disjoint subsets of) the repository,
 * it is necessary to link the WebNode objects between the machines, so
 * as to allow a WebSampler simulation to span multiple machines.
 * A LeafNode is an alias for a WebNode stored on another machine, and
 * linked to by a WebNode on the current machine. The data stored in 
 * a LeafNode is only sufficient to allow the corresponding WebNode to 
 * be found.
 * The class inherits memory management from MemoryPooled<T>.
 */
class LeafNode: public MemoryPooled<LeafNodeStruct> {
 public:

  LeafNode(uint32 idno);

  uint32 ID() 
    { return data.id; }
  uint16 Date() 
    { return data.date; }

  void SetDate(uint16 adate)
    { data.date = adate; }

  void ClearOccupationCount() 
    { 
      data.occupation_count = 0;
    }
  uint32 OccupationCount() 
    { return data.occupation_count; }
  void IncrementOccupationCount() 
    { data.occupation_count++; }

};

#endif
