/** @file urlfilter.cc */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "urlfilter.h"
#include <cassert>
#include <ctype.h>
#include <iostream>


URLFilter :: URLFilter(bool removesuf) {
  flags.remove_html_suffix = removesuf;
}


/**
 * this function takes a standardized url ( see NormalizeURLPath() )
 * and removes the trailing string /index.htm(l)
 * This is used to compactify the string before adding
 * it to the Trie (in a trie, common prefixes are harmless, but
 * common suffixes waste space) In case the remove_html_suffix
 * flag is set, other common html endings are also tokenized to
 * reduce space requirements.
 *
 * Note that this operation is irreversible (we cannot
 * reinsert the suffix /index.html reliably in all cases).
 * The string anurl is not modified.
 */
const char* URLFilter :: DeindexURL(const char *anurl) {
  // make a copy of the string
  assert(strlen(anurl) < STRINGBUF_LEN2);
  strcpy(deindex_scratchbuf,anurl);

  char *p = strstr(deindex_scratchbuf, "/index.htm");
  if( p && ((p[10] == 0) || (p[11] == 0)) ) {
    *p = 0;
  } else if( flags.remove_html_suffix ) { 
    p = strstr(deindex_scratchbuf, ".htm");
    if( p ) {
      if(p[4] == 0) { // ends in .htm
	p[0] = 31; // nonprintable
	p[1] = 0;
      } else if(p[5] == 0) { // ends in .html
	p[0] = 30;
	p[1] = 0;
      }
    } else { 
      p = strstr(deindex_scratchbuf, ".shtm");
      if( p ) {
	if(p[5] == 0) { // ends in .shtm
	  p[0] = 29;
	  p[1] = 0;
	} else if(p[6] == 0) { // ends in .shtml
	  p[0] = 28;
	  p[1] = 0;
	}
      } else { 
	p = strstr(deindex_scratchbuf, ".asp");
	if( p ) {
	  if(p[4] == 0) { // ends in .asp
	    p[0] = 27;
	    p[1] = 0;
	  } else {
	    p = strstr(deindex_scratchbuf, ".php");
	    if( p ) {
	      if(p[4] == 0) { // ends in .php
		p[0] = 26;
		p[1] = 0;
	      }
	    }
	  }
	}
      }
    }
  }
  return deindex_scratchbuf;
}

/**
 * This function compresses a URL, whose characters are 
 * guaranteed to fit within seven bits, and removes all
 * the forward slashes, which are the most commonly used
 * character. Everytime a slash is removed, the *preceding*
 * charater has its eight bit set. A slash is not removed
 * if the previous character already has its eight bit set.
 *
 * The compressed URL is always located in the special buffer
 * comp_scratchbuf[]. The string anurl is not modified
 */ 
const char* URLFilter :: CompressURL(const char *anurl) {
  const char *p = anurl; 
  char *q = comp_scratchbuf;
  while( *p ) {
    if( p[1] == '/' ) {
      *q++ = (*p | SLASHBIT); // print *p with slashbit
      p++;
    } else {
      *q++ = *p; // just print *p
    }
    p++;
  }
  *q = 0;
  return anurl;
}

char *stringreverse(char *str)
{
  register int max,high,low;

  max=strlen(str);
  high=max-1;
  low=0;

  while (low < high)
    {
    str[max] = str[low];
    str[low++] = str[high];
    str[high--] = str[max];
    }

  str[max] = 0;
  return str;
}

/// Decomposes a URL into its components for analysis.
/**
 * Each of the supplied buffers must be STRINGBUF_LEN1 long.
 * This function does not modify anurl.
 *
 * If flags.rearrange_components is true, the network location
 * and file path are rearranged so that the suffix is placed first.
 */
void URLFilter :: ParseURL(const char *anurl, char *schemebuf, 
			      char *netlocbuf, char *querybuf,
			      char *paramsbuf, char *pathbuf) {
  
  assert( strlen(anurl) <= STRINGBUF_LEN1 );

  strcpy(parse_scratchbuf, anurl);
  
  char *q;

  // get rid of the fragment if it's there
  char *p = strchr(parse_scratchbuf, '#');
  if( p ) { *p = 0; }

  // now get scheme
  p = strchr(parse_scratchbuf, ':');
  if( p ) {
    int l = p - parse_scratchbuf;
    strncpy(schemebuf, parse_scratchbuf, l);
    schemebuf[l] = 0;
    for(int k = 0; k < l; k++) {
      if( !isalnum(schemebuf[k]) && !(schemebuf[k] == '+') &&
	  !(schemebuf[k] == '.') && !(schemebuf[k] == '-') ) { // illegal
	schemebuf[0] = 0;
      }
      schemebuf[k] = tolower(schemebuf[k]);
    }
    q = p + 1; //strcpy(parse_scratchbuf2, p + 1); // rid of scheme
  } else { 
    schemebuf[0] = 0;
    q = parse_scratchbuf; //strcpy(parse_scratchbuf2, parse_scratchbuf1); // rid of scheme
  }

  // now get network location
  if( (q[0] == '/') && (q[1] == '/') ) { // double slash
    p = strchr(q + 2, '/');
    if( p ) {
      int l = p - q - 2;
      strncpy(netlocbuf, q + 2, l);
      netlocbuf[l] = 0;
      q = p;
      // strcpy(parse_scratchbuf3, p);
    } else {
      strcpy(netlocbuf, q + 2);
      q = parse_scratchbuf + strlen(parse_scratchbuf);
      assert( *q == 0 );
    }
  } else { // no double slash found
    netlocbuf[0] = 0;
    // q is still good
  }
  // we convert netloc to lowercase - 
  // could be problematic if login/passwords are used though
  for(char* cp = netlocbuf; *cp; cp++) { *cp = tolower(*cp); }

  // remove query string if present
  p = strchr(q, '?');
  if( p ) {
    strcpy(querybuf, p);
    *p = 0;
  } else {
    querybuf[0] = 0;
  }

  // remove parameters if present
  p = strchr(q, ';');
  if( p ) {
    strcpy(paramsbuf, p);
    *p = 0;
  } else {
    paramsbuf[0] = 0;
  }

  // what's left is the path
  strcpy(pathbuf, q);
}



/// Fixes the path in case document is of type index.html 
/**
 * This function maps paths of the form
 * xxx/
 * xxx/index.htm     to the standard  xxx/index.html
 * xxx/index.html    
 *
 * WARNING: This function modifies the string apath. It is 
 * assumed that apath has STRINGBUF_LEN1 storage available
 */
void URLFilter :: NormalizeURLPath(char *apath) {
  int j = strlen(apath) - 1;
  if( j > 0 ) {
    if( apath[j] == '/' ) {
      if( j < (STRINGBUF_LEN1 - 11) )
	strcat(apath, "index.html");
    } else if( apath[j] == 'm' ) { 
      // is it index.htm, then append 'l'
      if( (j >= 8) && 
	  (strncasecmp(apath + j - 8, "index.ht", 8) == 0) &&
	  (j < (STRINGBUF_LEN1 - 2)) ) {
	apath[++j] = 'l';
	apath[++j] = 0;
      }
    }
  } else { // path is empty
    strcpy(apath, "/index.html");
  }
}


/// classifies a file according to its extension
ContentType URLFilter :: ClassifyURLPath(const char *path) {
  ContentType ctype;
  int pathlen = strlen(path);

  if( (pathlen > 5) && (strncasecmp(path + pathlen - 5, ".html", 5) == 0) ) {
    ctype = CONTENT_TEXT_HTML;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".htm", 4) == 0) ) {
    ctype = CONTENT_TEXT_HTML;
  } else if( (pathlen > 6) && (strncasecmp(path + pathlen - 6, ".shtml", 6) == 0) ) {
    ctype = CONTENT_TEXT_HTML;
  } else if( (pathlen > 7) && (strncasecmp(path + pathlen - 7, ".readme", 7) == 0) ) {
    ctype = CONTENT_TEXT_PLAIN;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".asp", 4) == 0) ) {
    ctype = CONTENT_TEXT_HTML;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".lha", 4) == 0) ) {
    ctype = CONTENT_APPLICATION_XGZIP;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".txt", 4) == 0) ) {
    ctype = CONTENT_TEXT_PLAIN;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".rtf", 4) == 0) ) {
    ctype = CONTENT_TEXT_RTF;
  } else if( (pathlen > 3) && (strncasecmp(path + pathlen - 3, ".ps", 3) == 0) ) {
    ctype = CONTENT_APPLICATION_POSTSCRIPT;
  } else if( (pathlen > 3) && (strncasecmp(path + pathlen - 3, ".gz", 3) == 0) ) {
    ctype = CONTENT_APPLICATION_XGZIP;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".doc", 4) == 0) ) {
    ctype = CONTENT_APPLICATION_MSWORD;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".ppt", 4) == 0) ) {
    ctype = CONTENT_APPLICATION_MS_POWERPOINT;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".jpg", 4) == 0) ) {
    ctype = CONTENT_IMAGE;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".zip", 4) == 0) ) {
    ctype = CONTENT_APPLICATION_XGZIP;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".gif", 4) == 0) ) {
    ctype = CONTENT_IMAGE;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".png", 4) == 0) ) {
    ctype = CONTENT_IMAGE;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".pdf", 4) == 0) ) {
    ctype = CONTENT_APPLICATION_PDF;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".php", 4) == 0) ) {
    ctype = CONTENT_TEXT_HTML;
  } else if( (pathlen > 5) && (strncasecmp(path + pathlen - 5, ".jpeg", 5) == 0) ) {
    ctype = CONTENT_IMAGE;
  } else if( (pathlen > 5) && (strncasecmp(path + pathlen - 5, ".shtm", 5) == 0) ) {
    ctype = CONTENT_TEXT_HTML;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".ram", 4) == 0) ) {
    ctype = CONTENT_AUDIO_MP3;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".mov", 4) == 0) ) {
    ctype = CONTENT_AUDIO_MP3;
  } else if( (pathlen > 4) && (strncasecmp(path + pathlen - 4, ".mpg", 4) == 0) ) {
    ctype = CONTENT_AUDIO_MP3;
  } else {
    // otherwise, we don't recognize the type
    ctype = CONTENT_GOOGLE_OTHER;
  }
  return ctype;
}


// note this function completes relative urls with the current
// document name. This is only possible if another_url is 
// already defined.
//
//
/**
 * This function formats anurl into a standard form. Its most important
 * use is as a completion mechanism for URL fragments as can be found
 * in anchor tags. The URL is completed relative to the baseurl, which
 * typically is the current document's url.
 *
 * The return value will always be a pointer to one of the scratch buffers
 * so you should copy the returned string before formatting another.
 */
const char * URLFilter :: FormatURL(const char *anurl, int anurl_len, 
				    URLComponents *baseurl, ContentType *foundtype) throw (domain_error) {
  assert(anurl_len > 0);

  // overflow possibility:
  // if hreflen is greater than 1024, then the url is likely machine
  // generated and contains some crypto key - we keep only the fist
  // 1024 chars, since the resulting string is most likely unique already
  if( anurl_len >= STRINGBUF_LEN1 ) {
    cerr << "warning: truncating url" << endl;
    anurl_len = STRINGBUF_LEN1; 
  }

  char *fullhref = NULL;;

  // copy and make null terminated
  // also make sure that the string is composed of valid characters
  // from RFC 2396, we shall accept all ASCII chars excluding only
  // 0x00-0x1F, 0x7F, which are unprintable. This allows some non-URI
  // characters through, but those are harmless. 
  int a;
  for(a = 0; a < anurl_len; a++) {
    if( (anurl[a] <= 0x1F) || (anurl[a] >= 0x7F) ) {
      break;
    } else {
      scratchbuf1[a] = anurl[a];
    }
  }
  scratchbuf1[a] = 0;
  
  // this replaces the existing contents of all scratchbufs
  ParseURL(scratchbuf1, scratchbuf0, scratchbuf3, scratchbuf5, scratchbuf6, scratchbuf2);
  // if path ends in / we append index.html
  NormalizeURLPath(scratchbuf2);
  // set the content type
  *foundtype = ClassifyURLPath(scratchbuf2);

  if( strlen(scratchbuf3) > 0 ) { // if network location defined

    if( scratchbuf0[0] ) {
      strcpy(scratchbuf1, scratchbuf0);
      strcat(scratchbuf1, "://");
    } else { // insert default protocol scheme
      strcpy(scratchbuf1, "http://");
    }
    strcat(scratchbuf1, scratchbuf3);
    strcat(scratchbuf1, scratchbuf2);
    strcat(scratchbuf1, scratchbuf6);
    strcat(scratchbuf1, scratchbuf5);
    fullhref = scratchbuf1;

  } else if( baseurl ) { // it's relative to our document's network location

    if( scratchbuf0[0] ) {
      strcpy(scratchbuf4, scratchbuf0);
      strcat(scratchbuf4, "://");
    } else { // insert default protocol scheme
      strcpy(scratchbuf4, "http://");
    }
    strcat(scratchbuf4, baseurl->netloc);
    strcat(scratchbuf4, "/");

    // the path may be absolute or relative
    if( scratchbuf2[0] == '/' ) { // absolute

      strcat(scratchbuf4, scratchbuf2 + 1);
      strcat(scratchbuf4, scratchbuf6);
      strcat(scratchbuf4, scratchbuf5);
      fullhref = scratchbuf4;

    } else { // relative path

      // if it's a script or an html file we remove everything
      // until the last /

      int r = strlen(baseurl->path) - 1;
      if( strlen(baseurl->query) || strlen(baseurl->params) ||
	  ((r > 5) && (strncasecmp(baseurl->path + r - 5, ".shtml",6) == 0)) ||
	  ((r > 4) && (strncasecmp(baseurl->path + r - 4, ".html",5) == 0)) ||
	  ((r > 3) && (strncasecmp(baseurl->path + r - 3, ".htm",4) == 0)) ) {
	
	char *s = strrchr(baseurl->path, '/');
	assert(s);
	char *t = scratchbuf4 + strlen(scratchbuf4);
	char *u = baseurl->path + 1;
	while( u <= s ) {
	  *t++ = *u++;
	}
	*t = 0;
	// now ends with a slash

      } else { // neither script nor html, assume it's a bona fide directory

	if( baseurl->path[0] == '/' ) {
	  strcat(scratchbuf4, baseurl->path + 1);
	} else {
	  strcat(scratchbuf4, baseurl->path);
	}

	// append trailing slash if needed
	int l = strlen(scratchbuf4) - 1;
	if( (l >= 0) && (scratchbuf4[l] != '/') ) {
	  scratchbuf4[++l] = '/';
	  scratchbuf4[++l] = 0;
	} else if( l < 0) { 
	  scratchbuf4[0] = '/';
	  scratchbuf4[1] = 0;
	}
	// now ends with a slash
      }

       // now we can append the actual relative path 
      // note this doesnt start with /
      strcat(scratchbuf4, scratchbuf2);
      // now remove ../ patterns
      char *p;
      while( (p = strstr(scratchbuf4, "../")) ) {
	p[0] = p[1] = p[2] = 0x80;
	while( *p != '/' ) { *p-- = 0x80; }
	*p-- = 0x80;
	while( *p != '/' ) { *p-- = 0x80; }
      }

      // next remove ./ patterns
      while( (p = strstr(scratchbuf4, "./")) ) {
	p[0] = p[1] = 0x80;
	while( *p != '/' ) { *p-- = 0x80; }
      }

      // and clean up
      char *q = scratchbuf3; 
      p = scratchbuf4;
      while( *p ) { 
	if( *p & 0x80 ) {
	  p++;
	} else {
	  *q++ = *p++;
	}
      }
      *q = 0;

      strcat(scratchbuf3, scratchbuf6);
      strcat(scratchbuf3, scratchbuf5);

      fullhref = scratchbuf3;
    }
  } else { // relative url but baseurl == NULL
    cerr << "error: relative url encountered but baseurl == NULL" << endl;
    throw domain_error("");
  }

  // extra bit of code: some people use file: when they mean
  // http:, so here we replace the file scheme with http
  if( (strncasecmp(fullhref, "file", 4) == 0) 
    && strstr(fullhref,"htm") ) {
    fullhref[0] = 'h';
    fullhref[1] = 't';
    fullhref[2] = 't';
    fullhref[3] = 'p';
  }

  NormalizeURLPath(fullhref);
  // DEBUG cout << " * " << fullhref << endl;
 
  assert( strlen(fullhref) < STRINGBUF_LEN2 );

  return fullhref;
}

#ifdef UNIT_TEST

#define MAKE_TEST_STATEMENT(x) \
do { \
  cout << "testing: " #x << endl; \
  cout << "output: " << endl; \
  (x); \
} while(false)

#define MAKE_TEST(x,y) \
do { \
   cout << argv[0] << ": " << (((x) == (y)) ? "PASSED" : "FAILED") << " " #x <<endl; \
   cout << "output: " << (x) << endl; \
} while(false)

#define MAKE_TEST_STRING(x,y) \
do { \
   cout << argv[0] << ": " << ((strcmp((x),(y)) == 0) ? "PASSED" : "FAILED") << " " #y <<endl; \
   cout << "output: " << (x) << endl; \
} while(false)

int main(int argc, char** argv) {

  const char* strings[] = {
    "http://www.somewhere.edu/directory/",
    "http://www.site.com/default.htm",
    "file://www.buffalo.edu/admin/one.html",
    "../folder/special.shtm"
  };

  URLFilter *uf = new URLFilter(false);
  URLComponents base;
  ContentType contype;

  const char* result;
  
  cout << (result = uf->FormatURL(strings[0], strlen(strings[0]), NULL, &contype)) << endl;
  MAKE_TEST_STRING(result,"http://www.somewhere.edu/directory/index.html");

  cout << (result = uf->FormatURL(strings[1], strlen(strings[1]), NULL, &contype)) << endl;
  MAKE_TEST_STRING(result,"http://www.site.com/default.htm");
  MAKE_TEST(contype,CONTENT_TEXT_HTML);

  cout << (result = uf->FormatURL(strings[2], strlen(strings[2]), NULL, &contype)) << endl;
  MAKE_TEST_STRING(result,"http://www.buffalo.edu/admin/one.html");

  uf->ParseURL(strings[0], base.scheme, base.netloc, base.query, base.params, base.path);
  cout << (result = uf->FormatURL(strings[3], strlen(strings[3]), &base, &contype)) << endl;
  MAKE_TEST_STRING(result,"http://www.somewhere.edu/folder/special.shtm");

  char bug[10]; strcpy(bug,"hello");
  MAKE_TEST_STRING(stringreverse(bug), "olleh");

//   cout << (result = uf->FormatURL(strings[0], strlen(strings[0]), &base, &contype)) << endl;
//   MAKE_TEST_STRING(result,"http://www.somewhere.edu/directory/index.html");

//   cout << (result = uf->FormatURL(strings[0], strlen(strings[0]), &base, &contype)) << endl;
//   MAKE_TEST_STRING(result,"http://www.somewhere.edu/directory/index.html");

//   cout << (result = uf->FormatURL(strings[0], strlen(strings[0]), &base, &contype)) << endl;
//   MAKE_TEST_STRING(result,"http://www.somewhere.edu/directory/index.html");

}
#endif
